pub struct RleAnalysis {
    pub original_size: usize,
    pub encoded_unescaped_size_estimate: usize,
    pub smallest_least_common_byte: u8,
    pub count: [u64; 256]
}

pub fn rle_analyse(input: &[u8]) -> RleAnalysis {
    let mut original_size = 0;
    let mut encoded_unescaped_size_estimate = 0;

    let mut count: [u64; 256] = [0; 256];
    let mut last_byte: Option<u8> = None;
    for &b in input {
        if Some(b) != last_byte {
            encoded_unescaped_size_estimate += 2;
            last_byte = Some(b);
        }
        original_size += 1;
        count[b as usize] += 1;
    }

    RleAnalysis {
        original_size,
        encoded_unescaped_size_estimate,
        smallest_least_common_byte: count.iter().enumerate()
            .fold((0u8, <u64>::max_value()), |(byte, ammount), (new_byte, new_ammount)| {
                if *new_ammount < ammount {
                    (new_byte as u8, *new_ammount)
                } else {
                    (byte, ammount)
                }
            }).0,
        count
    }
}

pub fn rle_encode(input: &[u8], escape_byte: Option<u8>) -> Vec<u8> {
    let mut encoded = Vec::new(); // TODO Many allocations :(
    let mut last_byte: Option<u8> = None;
    let mut current_len: u8 = 0;

    if let Some(eb) = escape_byte {
        encoded.push(eb);
    }

    for &b in input {
        if let Some(escape_byte_u) = escape_byte {
            if let Some(last_byte_u) = last_byte {
                if last_byte_u != b || current_len == 255 {
                    if current_len > 2 {
                        encoded.push(escape_byte_u);
                        encoded.push(current_len);
                        encoded.push(last_byte_u);
                    } else if current_len == 2 {
                        if last_byte_u == escape_byte_u {
                            encoded.push(escape_byte_u);
                            encoded.push(2);
                            encoded.push(last_byte_u);
                        } else {
                            encoded.push(last_byte_u);
                            encoded.push(last_byte_u);
                        }
                    } else {
                        if last_byte_u == escape_byte_u {
                            encoded.push(escape_byte_u);
                            encoded.push(1);
                            encoded.push(last_byte_u);
                        } else {
                            encoded.push(last_byte_u);
                        }
                    }
                    last_byte = Some(b);
                    current_len = 1;
                } else {
                    current_len += 1;
                }
            } else {
                last_byte = Some(b);
                current_len = 1;
            }
        } else {
            // Without escapes
            if let Some(last_byte_u) = last_byte {
                if last_byte_u != b || current_len == 255 {
                    encoded.push(current_len);
                    encoded.push(last_byte_u);
                    last_byte = Some(b);
                    current_len = 1;
                } else {
                    current_len += 1;
                }
            } else {
                last_byte = Some(b);
                current_len = 1;
            }
        }
    }

    // TODO Deduplicate code
    if let Some(escape_byte_u) = escape_byte {
        if let Some(last_byte_u) = last_byte {
            if current_len > 2 {
                encoded.push(escape_byte_u);
                encoded.push(current_len);
                encoded.push(last_byte_u);
            } else if current_len == 2 {
                if last_byte_u == escape_byte_u {
                    encoded.push(escape_byte_u);
                    encoded.push(2);
                    encoded.push(last_byte_u);
                } else {
                    encoded.push(last_byte_u);
                        encoded.push(last_byte_u);
                }
            } else {
                if last_byte_u == escape_byte_u {
                    encoded.push(escape_byte_u);
                    encoded.push(1);
                    encoded.push(last_byte_u);
                } else {
                    encoded.push(last_byte_u);
                }
            }
        }
    } else {
        // Without escapes
        if let Some(last_byte_u) = last_byte {
            encoded.push(current_len);
            encoded.push(last_byte_u);
        }
    }

    encoded
}

pub fn rle_decode(input: &[u8], escaped: bool) -> Vec<u8> {
    let mut decodeed: Vec<u8> = Vec::new(); // TODO Many allocations :(

    let mut len: Option<u8> = None;
    if escaped {
        if let Some(&escape_byte) = input.get(0) {
            let mut in_escape = false;
            for &val in input[1..].iter() {
                if in_escape {
                    if let Some(l) = len {
                        for _ in 0..l {
                            decodeed.push(val);
                            len = None;
                            in_escape = false;
                        }
                    } else {
                        len = Some(val);
                    }
                } else {
                    if escape_byte == val {
                        in_escape = true;
                    } else {
                        decodeed.push(val);
                    }
                }
            }
        } else {
            // TODO Report malformed itterator
        }
    } else {
        // Without escapes
        for &val in input {
            if let Some(l) = len {
                for _ in 0..l {
                    decodeed.push(val);
                    len = None;
                }
            } else {
                len = Some(val);
            }
        }
    }

    // TODO Report possible remaining byte as error
    
    decodeed
}

#[cfg(test)]
mod tests {
    use super::*;

    mod encode {
        use super::*;
        
        #[test]
        fn escaped_1() {
            assert_eq!(rle_encode(&[1,1,1,1,2,3,4,5,5,6,6,6,6,6], Some(0)),
                       vec![0, 0,4,1, 2, 3, 4, 5,5, 0,5,6]);
        }
        
        #[test]
        fn escaped_2() {
            assert_eq!(rle_encode(&[1,1,9,1,1,2,3,4,5,5,5,9,9,6,6,6], Some(9)),
                       vec![9, 1,1, 9,1,9, 1,1, 2, 3, 4, 9,3,5, 9,2,9, 9,3,6]);
        }
        
        #[test]
        fn escaped_full_block_1() {
            assert_eq!(rle_encode(&[10; 259], Some(0)),
                       vec![0, 0,255,10, 0,4,10]);
        }

        
        #[test]
        fn unescaped_1() {
            assert_eq!(rle_encode(&[1,1,1,1,2,3,4,5,5,6,6,6,6,6], None),
                       vec![4,1, 1,2, 1,3, 1,4, 2,5, 5,6]);
        }

        #[test]
        fn unescaped_full_block_1() {
            assert_eq!(rle_encode(&[10; 259], None),
                       vec![255, 10, 4, 10]);
        }
    }

    mod decode {
        use super::*;

        #[test]
        fn escaped_1() {
            assert_eq!(rle_decode(&[0, 0,4,1, 2, 3, 4, 5,5, 0,5,6], true),
                       vec![1,1,1,1,2,3,4,5,5,6,6,6,6,6]);
        }

        #[test]
        fn escaped_full_block_1() {
            assert_eq!(rle_decode(&[0, 0,255,10, 0,4,10], true),
                       vec![10; 259]);
        }

        #[test]
        fn unescaped_1() {
            assert_eq!(rle_decode(&[4,1, 1,2, 1,3, 1,4, 2,5, 5,6], false),
                       vec![1,1,1,1,2,3,4,5,5,6,6,6,6,6]);
        }

        #[test]
        fn unescaped_full_block_1() {
            assert_eq!(rle_decode(&[255, 10, 4, 10], false),
                       vec![10; 259]);
        }
    }
}
